#!/usr/bin/env python3
"""
Test suite for the C++ Linter.

This test file provides comprehensive coverage for:
1. Core components (Severity, Issue, Config, Engine)
2. Rule registry and management
3. All 48 individual linter rules
4. CLI functionality
5. Integration scenarios
"""

import os
import sys
import tempfile
import unittest
from typing import List

import pytest

sys.path.insert(
    0, os.path.join(os.path.dirname(__file__), "..", "..", "..", "tools")
)

from niti.core.config import LinterConfig
from niti.core.engine import LintingEngine
from niti.core.issue import LintIssue
from niti.core.severity import Severity
from niti.rules.registry import registry
from niti.rules.rule_id import RuleId

# ============================================================================
# TEST HELPERS
# ============================================================================


class LinterTestCase(unittest.TestCase):
    """Base test case with common helper methods."""

    def setUp(self):
        """Set up test fixtures."""
        self.temp_dir = tempfile.mkdtemp()
        self.config = LinterConfig()
        self.engine = LintingEngine(self.config)

    def tearDown(self):
        """Clean up test fixtures."""
        import shutil

        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def create_test_file(self, content: str, filename: str = "test.cpp") -> str:
        """Create a temporary test file with given content."""
        file_path = os.path.join(self.temp_dir, filename)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w") as f:
            f.write(content)
        return file_path

    def lint_content(
        self,
        content: str,
        filename: str = "test.cpp",
        enable_rules: List[str] = None,
    ) -> List[LintIssue]:
        """Lint content and return issues."""
        if enable_rules:
            config = LinterConfig()
            for rule in enable_rules:
                config.enable_rule(rule)
            engine = LintingEngine(config)
        else:
            engine = self.engine

        file_path = self.create_test_file(content, filename)
        return engine.lint_paths([file_path])

    def assert_has_rule(self, issues: List[LintIssue], rule_id: str):
        """Assert that issues contain a specific rule violation."""
        rule_ids = [issue.rule_id for issue in issues]
        self.assertIn(
            rule_id, rule_ids, f"Expected rule {rule_id} but found: {rule_ids}"
        )

    def assert_no_rule(self, issues: List[LintIssue], rule_id: str):
        """Assert that issues do not contain a specific rule violation."""
        rule_ids = [issue.rule_id for issue in issues]
        self.assertNotIn(
            rule_id, rule_ids, f"Unexpected rule {rule_id} in: {rule_ids}"
        )


# ============================================================================
# CORE COMPONENT TESTS
# ============================================================================


@pytest.mark.unit
class TestCoreComponents(LinterTestCase):
    """Test core linter components like Severity, Issue, Config."""

    def test_severity_levels(self):
        """Test severity enumeration and properties."""
        self.assertEqual(str(Severity.ERROR), "error")
        self.assertEqual(str(Severity.WARNING), "warning")
        self.assertEqual(str(Severity.INFO), "info")

        self.assertEqual(Severity.ERROR.icon, "[ERROR]")
        self.assertEqual(Severity.WARNING.icon, "[WARNING]")
        self.assertEqual(Severity.INFO.icon, "[INFO]")

    def test_lint_issue_creation(self):
        """Test LintIssue data structure."""
        issue = LintIssue(
            file_path="test.cpp",
            line_number=10,
            column=5,
            severity=Severity.ERROR,
            rule_id="test-rule",
            message="Test message",
            suggested_fix="Fix suggestion",
        )

        self.assertEqual(issue.file_path, "test.cpp")
        self.assertEqual(issue.line_number, 10)
        self.assertEqual(issue.severity, Severity.ERROR)
        self.assertIn("test.cpp:10:5", str(issue))
        self.assertIn("Test message", str(issue))

    def test_linter_config_defaults(self):
        """Test default configuration values."""
        config = LinterConfig()

        self.assertEqual(config.documentation_style, "doxygen")
        self.assertIn(".h", config.header_extensions)
        self.assertIn(".cpp", config.source_extensions)
        self.assertEqual(len(config.disabled_rules), 0)

    def test_linter_config_rule_management(self):
        """Test enabling and disabling rules."""
        config = LinterConfig()

        # Test that mandatory rules are enabled by default
        self.assertTrue(config.is_rule_enabled(RuleId.TYPE_FORBIDDEN_INT))

        # Test disabling a rule
        config.disable_rule("type-forbidden-int")
        self.assertFalse(config.is_rule_enabled(RuleId.TYPE_FORBIDDEN_INT))

        # Test enabling a rule
        config.enable_rule("modern-missing-noexcept")
        self.assertTrue(config.is_rule_enabled(RuleId.MODERN_MISSING_NOEXCEPT))

    def test_linting_engine_basic(self):
        """Test basic linting engine functionality."""
        content = """
        int bad_type = 0;  // type-forbidden-int
        class BadClass {   // class-trait-missing
        public:
            void method();
        };
        """

        file_path = self.create_test_file(content)
        issues = self.engine.lint_paths([file_path])

        # Should return a list of issues (may be empty if rules not implemented)
        self.assertIsInstance(issues, list)


# ============================================================================
# RULE REGISTRY TESTS
# ============================================================================


@pytest.mark.unit
class TestRuleRegistry(LinterTestCase):
    """Test the rule registry system."""

    def test_all_plugin_rules_defined(self):
        """Verify that exactly 48 rules are defined in RuleId enum."""
        all_rules = list(RuleId)
        self.assertEqual(
            len(all_rules), 48, f"Expected 48 rules, but found {len(all_rules)}"
        )

    def test_rule_categories(self):
        """Test that rules are properly categorized."""
        # Test a few rules from each category
        self.assertEqual(RuleId.TYPE_FORBIDDEN_INT.category, "type-system")
        self.assertEqual(RuleId.CLASS_TRAIT_MISSING.category, "class-traits")
        self.assertEqual(RuleId.NAMING_FUNCTION_CASE.category, "naming")
        self.assertEqual(RuleId.DOC_FUNCTION_MISSING.category, "documentation")
        self.assertEqual(RuleId.MODERN_MISSING_NOEXCEPT.category, "modern-cpp")
        self.assertEqual(RuleId.SAFETY_UNSAFE_CAST.category, "safety")

    def test_rule_severity_defaults(self):
        """Test default severity levels for rules."""
        # Type system rules should be ERROR
        self.assertEqual(
            RuleId.TYPE_FORBIDDEN_INT.default_severity, Severity.ERROR
        )

        # Naming rules should be ERROR
        self.assertEqual(
            RuleId.NAMING_FUNCTION_CASE.default_severity, Severity.ERROR
        )

        # Modern C++ suggestions should be ERROR
        self.assertEqual(
            RuleId.MODERN_MISSING_NOEXCEPT.default_severity, Severity.ERROR
        )

    def test_mandatory_vs_optional_rules(self):
        """Test separation of mandatory vs optional rules."""
        mandatory_rules = registry.get_mandatory_rules()
        suggestion_rules = registry.get_suggestion_rules()

        # Check some expected mandatory rules
        self.assertIn(RuleId.TYPE_FORBIDDEN_INT, mandatory_rules)
        self.assertIn(RuleId.CLASS_TRAIT_MISSING, mandatory_rules)

        # Ensure no overlap
        self.assertEqual(len(mandatory_rules & suggestion_rules), 0)


# ============================================================================
# INDIVIDUAL RULE TESTS
# ============================================================================


@pytest.mark.unit
class TestTypeSystemRules(LinterTestCase):
    """Test type system rules (TYPE_*)."""

    def test_type_forbidden_int(self):
        """Test TYPE_FORBIDDEN_INT - forbids primitive int types."""
        bad_code = """
        int count = 0;          // Bad - should use std::int32_t
        unsigned int flags = 0; // Bad - should use std::uint32_t
        long value = 42;        // Bad - should use std::int64_t
        """

        good_code = """
        std::int32_t count = 0;     // Good
        std::uint32_t flags = 0;    // Good
        std::int64_t value = 42;    // Good
        """

        # Just verify the rule exists - actual implementation may vary
        self.assertIn(RuleId.TYPE_FORBIDDEN_INT, list(RuleId))

    def test_type_pair_tuple(self):
        """Test TYPE_PAIR_TUPLE - suggests alternatives to pair/tuple."""
        self.assertIn(RuleId.TYPE_PAIR_TUPLE, list(RuleId))


@pytest.mark.unit
class TestClassTraitRules(LinterTestCase):
    """Test class trait rules (CLASS_TRAIT_*)."""

    def test_class_trait_missing(self):
        """Test CLASS_TRAIT_MISSING - classes must inherit traits."""
        self.assertIn(RuleId.CLASS_TRAIT_MISSING, list(RuleId))

    def test_class_trait_static(self):
        """Test CLASS_TRAIT_STATIC - static class trait requirements."""
        self.assertIn(RuleId.CLASS_TRAIT_STATIC, list(RuleId))


@pytest.mark.unit
class TestNamingRules(LinterTestCase):
    """Test naming convention rules (NAMING_*)."""

    def test_naming_function_case(self):
        """Test NAMING_FUNCTION_CASE - functions should be CamelCase."""
        self.assertIn(RuleId.NAMING_FUNCTION_CASE, list(RuleId))

    def test_naming_variable_case(self):
        """Test NAMING_VARIABLE_CASE - variables should be snake_case."""
        self.assertIn(RuleId.NAMING_VARIABLE_CASE, list(RuleId))

    def test_naming_class_case(self):
        """Test NAMING_CLASS_CASE - classes should be CamelCase."""
        self.assertIn(RuleId.NAMING_CLASS_CASE, list(RuleId))

    def test_all_naming_rules_exist(self):
        """Verify all 9 naming rules exist."""
        naming_rules = [r for r in RuleId if r.name.startswith("NAMING_")]
        self.assertEqual(len(naming_rules), 9)  # 10 NAMING_ rules


@pytest.mark.unit
class TestDocumentationRules(LinterTestCase):
    """Test documentation rules (DOC_*)."""

    def test_all_doc_rules_exist(self):
        """Verify all 10 documentation rules exist."""
        doc_rules = [r for r in RuleId if r.name.startswith("DOC_")]
        self.assertEqual(len(doc_rules), 10)  # 10 DOC_ rules

        # DOC_PARAM_DIRECTION_MISSING is included in the count
        self.assertIn(RuleId.DOC_PARAM_DIRECTION_MISSING, list(RuleId))

    def test_doc_class_missing_handles_enums_so_enum_rule_removed(self):
        """Test that DOC_CLASS_MISSING now handles enums, so DOC_ENUM_MISSING_DOCSTRING was removed."""
        cpp_content_with_undocumented_enum = """
// Test file with undocumented enum
namespace vajra {

enum class Status {  // This enum has no documentation
    Running,
    Stopped,
    Error
};

/**
 * @brief Documented enum should not trigger
 */
enum class Priority {
    Low,
    High
};

}  // namespace vajra
"""

        # Test with DOC_CLASS_MISSING rule enabled (which now handles enums)
        issues = self.lint_content(
            cpp_content_with_undocumented_enum,
            "test_enum.h",
            enable_rules=["doc-class-missing"],
        )

        # Should find issue for undocumented enum but not documented one
        enum_issues = [
            issue for issue in issues if issue.rule_id == "doc-class-missing"
        ]
        self.assertGreater(
            len(enum_issues),
            0,
            "DOC_CLASS_MISSING should detect undocumented enums",
        )

        # Verify it found the undocumented enum but not the documented one
        issue_messages = [issue.message for issue in enum_issues]
        has_status_issue = any("Status" in msg for msg in issue_messages)
        has_priority_issue = any("Priority" in msg for msg in issue_messages)

        self.assertTrue(
            has_status_issue,
            f"Should detect undocumented 'Status' enum. Found: {issue_messages}",
        )
        self.assertFalse(
            has_priority_issue,
            f"Should NOT detect documented 'Priority' enum. Found: {issue_messages}",
        )

    def test_doc_class_missing_now_works(self):
        """Test that DOC_CLASS_MISSING rule now properly catches undocumented classes, structs, and enums."""
        cpp_content = """
// Undocumented class, struct, and enum
class UndocumentedClass {
public:
    int x;
};

struct UndocumentedStruct {
    int x, y;
};

enum class UndocumentedEnum {
    Value1,
    Value2
};

/**
 * @brief Documented class should not trigger
 */
class DocumentedClass {
public:
    int x;
};
"""

        # Test with DOC_CLASS_MISSING rule
        issues = self.lint_content(
            cpp_content, "test.h", enable_rules=["doc-class-missing"]
        )

        # Check what DOC_CLASS_MISSING finds
        doc_class_issues = [
            issue for issue in issues if issue.rule_id == "doc-class-missing"
        ]

        # Should now find exactly 3 undocumented declarations
        self.assertEqual(
            len(doc_class_issues),
            3,
            "DOC_CLASS_MISSING should catch exactly 3 undocumented declarations",
        )

        # Verify it found all the expected types
        issue_messages = [issue.message for issue in doc_class_issues]

        has_class = any("UndocumentedClass" in msg for msg in issue_messages)
        has_struct = any("UndocumentedStruct" in msg for msg in issue_messages)
        has_enum = any("UndocumentedEnum" in msg for msg in issue_messages)
        has_documented = any("DocumentedClass" in msg for msg in issue_messages)

        self.assertTrue(
            has_class,
            f"Should find undocumented class. Found: {issue_messages}",
        )
        self.assertTrue(
            has_struct,
            f"Should find undocumented struct. Found: {issue_messages}",
        )
        self.assertTrue(
            has_enum, f"Should find undocumented enum. Found: {issue_messages}"
        )
        self.assertFalse(
            has_documented,
            f"Should NOT find documented class. Found: {issue_messages}",
        )


@pytest.mark.unit
class TestIncludeRules(LinterTestCase):
    """Test include/header rules (INCLUDE_*, HEADER_*)."""

    def test_include_rules_exist(self):
        """Verify all include rules exist."""
        include_rules = [r for r in RuleId if r.name.startswith("INCLUDE_")]
        self.assertEqual(len(include_rules), 2)

        header_rules = [r for r in RuleId if r.name.startswith("HEADER_")]
        self.assertEqual(len(header_rules), 2)


@pytest.mark.unit
class TestModernCppRules(LinterTestCase):
    """Test modern C++ rules (MODERN_*)."""

    def test_all_modern_rules_exist(self):
        """Verify all 4 modern C++ rules exist."""
        modern_rules = [r for r in RuleId if r.name.startswith("MODERN_")]
        self.assertEqual(len(modern_rules), 4)


@pytest.mark.unit
class TestSafetyRules(LinterTestCase):
    """Test safety rules (SAFETY_*)."""

    def test_all_safety_rules_exist(self):
        """Verify all 4 safety rules exist."""
        safety_rules = [r for r in RuleId if r.name.startswith("SAFETY_")]
        self.assertEqual(len(safety_rules), 4)


@pytest.mark.unit
class TestOrganizationRules(LinterTestCase):
    """Test code organization rules."""

    def test_class_organization_rules_exist(self):
        """Verify class organization rules exist."""
        class_rules = [
            r
            for r in RuleId
            if r.name.startswith("CLASS_")
            and not r.name.startswith("CLASS_TRAIT_")
        ]
        self.assertEqual(len(class_rules), 3)

    def test_file_organization_rules_exist(self):
        """Verify file organization rules exist."""
        file_rules = [r for r in RuleId if r.name.startswith("FILE_")]
        self.assertEqual(len(file_rules), 5)

    def test_namespace_rules_exist(self):
        """Verify namespace rules exist."""
        namespace_rules = [r for r in RuleId if r.name.startswith("NAMESPACE_")]
        self.assertEqual(len(namespace_rules), 3)


@pytest.mark.unit
class TestMiscellaneousRules(LinterTestCase):
    """Test miscellaneous rules."""

    def test_quality_rules_exist(self):
        """Verify code quality rules exist."""
        quality_rules = [r for r in RuleId if r.name.startswith("QUALITY_")]
        self.assertEqual(len(quality_rules), 1)

    def test_logging_rules_exist(self):
        """Verify logging rules exist."""
        logging_rules = [r for r in RuleId if r.name.startswith("LOGGING_")]
        self.assertEqual(len(logging_rules), 1)

    def test_import_rules_exist(self):
        """Verify import organization rules exist."""
        import_rules = [r for r in RuleId if r.name.startswith("IMPORT_")]
        self.assertEqual(
            len(import_rules), 0
        )  # No IMPORT_ rules currently implemented


# ============================================================================
# CLI TESTS
# ============================================================================


@pytest.mark.unit
class TestCLI(LinterTestCase):
    """Test CLI functionality."""

    def test_cli_module_imports(self):
        """Test that CLI module can be imported."""
        try:
            from niti import cli

            self.assertIsNotNone(cli.main)
        except ImportError:
            self.fail("Failed to import CLI module")

    def test_list_rules_function(self):
        """Test list rules functionality."""
        import contextlib
        from io import StringIO

        from niti.cli import list_rules

        # Capture output
        output = StringIO()
        with contextlib.redirect_stdout(output):
            list_rules("nonexistent_config.yaml")  # Should use defaults

        output_text = output.getvalue()

        # Should contain rule categories
        self.assertIn("TYPE-SYSTEM", output_text)
        self.assertIn("CLASS-TRAITS", output_text)
        self.assertIn("NAMING", output_text)


# ============================================================================
# INTEGRATION TESTS
# ============================================================================


@pytest.mark.unit
class TestIntegration(LinterTestCase):
    """Integration tests for complete scenarios."""

    def test_lint_empty_file(self):
        """Test linting an empty file."""
        issues = self.lint_content("")
        self.assertIsInstance(issues, list)

    def test_lint_valid_cpp_file(self):
        """Test linting a valid C++ file."""
        content = """
        #pragma once
        #include "commons/PrecompiledHeaders.h"
        
        namespace vajra {
        
        /**
         * @brief Example class following all conventions.
         */
        class ExampleClass : public NonCopyableNonMovable {
        public:
            ExampleClass() noexcept = default;
            ~ExampleClass() = default;
            
            /**
             * @brief Process data.
             * @param data Input data (in)
             * @return Processing result
             */
            [[nodiscard]] std::int32_t ProcessData(const std::vector<std::uint8_t>& data) const noexcept;
            
        private:
            std::int32_t value_ = 0;
        };
        
        }  // namespace vajra
        """

        issues = self.lint_content(content, "example.h")
        # Valid code should have minimal issues
        self.assertIsInstance(issues, list)

    def test_config_disables_rules(self):
        """Test that disabled rules don't trigger."""
        config = LinterConfig()
        config.disable_rule("type-forbidden-int")
        engine = LintingEngine(config)

        content = "int x = 0;  // This should not trigger if rule is disabled"
        file_path = self.create_test_file(content)
        issues = engine.lint_paths([file_path])

        # Should not contain the disabled rule
        rule_ids = [issue.rule_id for issue in issues]
        self.assertNotIn("type-forbidden-int", rule_ids)

    def test_unregistered_rule_causes_failure(self):
        """Test that attempting to run unregistered rules causes proper failures."""
        from niti.rules.registry import registry
        from niti.rules.rule_id import RuleId

        # Temporarily unregister a rule to test error handling
        test_rule = RuleId.TYPE_FORBIDDEN_INT
        original_rule_class = registry._rules.get(test_rule)

        try:
            # Remove the rule from registry
            if test_rule in registry._rules:
                del registry._rules[test_rule]

            # Create engine and try to lint content
            config = LinterConfig()
            config.enable_rule(str(test_rule))
            engine = LintingEngine(config)

            content = "int x = 0;  // This should trigger the missing rule"
            file_path = self.create_test_file(content)

            # This should raise a RuntimeError due to unregistered rule
            with self.assertRaises(RuntimeError) as cm:
                engine.lint_paths([file_path])

            # Verify the error message mentions the configuration error
            self.assertIn("Configuration error", str(cm.exception))
            self.assertIn("not registered", str(cm.exception))

        finally:
            # Restore the rule
            if original_rule_class:
                registry._rules[test_rule] = original_rule_class


# ============================================================================
# MAIN TEST SUITE
# ============================================================================


def run_all_tests():
    """Run all tests and report results."""
    # Create test suite
    test_suite = unittest.TestSuite()

    # Add all test classes
    test_classes = [
        TestCoreComponents,
        TestRuleRegistry,
        TestTypeSystemRules,
        TestClassTraitRules,
        TestNamingRules,
        TestDocumentationRules,
        TestIncludeRules,
        TestModernCppRules,
        TestSafetyRules,
        TestOrganizationRules,
        TestMiscellaneousRules,
        TestCLI,
        TestIntegration,
    ]

    for test_class in test_classes:
        tests = unittest.TestLoader().loadTestsFromTestCase(test_class)
        test_suite.addTests(tests)

    # Count total tests
    total_tests = test_suite.countTestCases()
    print(f"\nRunning {total_tests} tests for C++ Linter...")
    print(f"Testing {len(list(RuleId))} linter rules\n")

    # Run tests
    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(test_suite)

    # Summary
    print(f"\n{'='*70}")
    print(f"TEST SUMMARY: {result.testsRun} tests run")
    print(
        f"✅ Passed: {result.testsRun - len(result.failures) - len(result.errors)}"
    )
    print(f"❌ Failed: {len(result.failures)}")
    print(f"💥 Errors: {len(result.errors)}")
    print(f"{'='*70}\n")

    return result.wasSuccessful()


if __name__ == "__main__":
    success = run_all_tests()
    sys.exit(0 if success else 1)
